# coding=utf-8
import pandas as pd
import re

from ClickSQL.clickhouse.ClickHouse import ClickHouseTableNode
from ClickSQL.errors import ClickHouseTableExistsError, ParameterTypeError


class CreateTableUtils(object):
    @staticmethod
    def _create_table_sql(db: str, table: str, dtypes_dict: dict,
                          order_by_key_cols: (list, tuple),
                          primary_key_cols=None, sample_expr=None,
                          engine_type: str = 'ReplacingMergeTree', partitions_expr=None,
                          settings="SETTINGS index_granularity = 8192", other=''):
        """

        :param db:
        :param table:
        :param dtypes_dict:
        :param order_by_key_cols:
        :param primary_key_cols:
        :param sample_expr:
        :param engine_type:
        :param partitions_expr:
        :param settings:
        :param other:
        :return:
        """
        """CREATE TABLE [IF NOT EXISTS] [db.]table_name [ON CLUSTER cluster]
                   (
                       name1 [type1] [DEFAULT|MATERIALIZED|ALIAS expr1],
                       name2 [type2] [DEFAULT|MATERIALIZED|ALIAS expr2],
                       ...
                   ) ENGINE = ReplacingMergeTree([ver])
                   [PARTITION BY expr]
                   [ORDER BY expr]
                   [PRIMARY KEY expr]
                   [SAMPLE BY expr]
                   [SETTINGS name=value, ...]"""

        if partitions_expr is None:
            partition_by_clause = ''
        else:
            partition_by_clause = f"PARTITION BY {partitions_expr}"

        if primary_key_cols is None:
            primary_by_clause = ''
        else:
            primary_key_expr = ','.join(primary_key_cols)
            primary_by_clause = f"PRIMARY BY ({primary_key_expr})"

        if sample_expr is None:
            sample_clause = ''
        else:
            sample_clause = sample_expr

        cols_def = ','.join([f"{name} {d_type}" for name, d_type in dtypes_dict.items()])

        main_body = f"CREATE TABLE IF NOT EXISTS {db}.{table} ( {cols_def} ) ENGINE = {engine_type}"

        ORDER_BY_CLAUSE = f"ORDER BY ( {','.join(order_by_key_cols)} )"

        cond_clause = f" {partition_by_clause} {ORDER_BY_CLAUSE} {primary_by_clause} {sample_clause} "

        base = f"{main_body} {cond_clause}  {other} {settings} "

        return base

    @staticmethod
    def _check_end_with_limit(string, pattern=r'[\s]+limit[\s]+[0-9]+$'):
        m = re.findall(pattern, string)
        if m is None or m == []:
            return False
        else:
            return True

    @staticmethod
    def translate_dtypes1_as_dtypes2(df: pd.DataFrame, src2target={'category': 'str'}):
        dtypes_series = df.dtypes
        for src, dest in src2target.items():
            if src in dtypes_series:
                category_cols = dtypes_series[dtypes_series == src].index
                for col in category_cols:
                    df[col] = df[col].astype(dest)
            else:
                pass
        return df

    @staticmethod
    def translate_dtypes_from_df(df: pd.DataFrame,
                                 translate_dtypes: dict = {'object': 'String', 'datetime64[ns]': 'Datetime'}):

        if hasattr(df, 'dtypes'):
            dtypes_series = df.dtypes.replace(translate_dtypes)
            return dtypes_series.map(lambda x: str(x).capitalize()).to_dict()
        elif hasattr(df, '_columns_') and 'type' in df._columns_ and 'name' in df._columns_:
            dtypes_series = df.set_index('name')['type'].replace(translate_dtypes)
            return dtypes_series.map(lambda x: str(x)).to_dict()
        else:
            raise ValueError(f'unknown df:{type(df)}')

    @classmethod
    def _create_table_from_df(cls, db: str, table: str, df: pd.DataFrame, key_cols: (list, tuple),
                              primary_key_cols=None, sample_expr=None,
                              engine_type: str = 'ReplacingMergeTree', extra_format_dict=None, partitions_expr=None,
                              settings="SETTINGS index_granularity = 8192",
                              other=''):

        df = cls.translate_dtypes1_as_dtypes2(df, src2target={'category': 'str'})
        cols = df.columns
        dtypes_dict = cls.translate_dtypes_from_df(df)
        if extra_format_dict is None:
            pass
        else:
            dtypes_dict.update(extra_format_dict)
        dtypes_dict = {k: v for k, v in dtypes_dict.items() if k in cols}
        base = cls._create_table_sql(db, table, dtypes_dict, key_cols,
                                     primary_key_cols=primary_key_cols,
                                     sample_expr=sample_expr,
                                     engine_type=engine_type,
                                     partitions_expr=partitions_expr,
                                     settings=settings,
                                     other=other
                                     )
        return base

    @classmethod
    def _obtain_describe_sql(cls, sql, pattern=r'[\s]+limit[\s]+[0-9]+$'):
        """
        detect limit end and obtain describe sql
        :param sql:
        :param pattern:
        :return:
        """
        limit_status = cls._check_end_with_limit(sql, pattern=pattern)
        if limit_status:
            describe_sql = f' describe({sql}) '
        else:
            describe_sql = f'describe ( {sql} limit 1)'

        return describe_sql


class ClickHouseTableNodeExt(ClickHouseTableNode, CreateTableUtils):
    def __init__(self, conn_str: (str, dict, None) = None, **kwarg):
        super(ClickHouseTableNodeExt, self).__init__(conn_str=conn_str, **kwarg)

    def create_view_if_not_exists(self, db: str, table: str, sql: str,
                                  key_cols: list,
                                  primary_key_cols=None, sample_expr=None,
                                  engine_type: str = 'view',
                                  extra_format_dict: (dict, None) = None, partitions_expr: (str, None) = None):
        if not self._check_exists(f'{db}.{table}', mode='table'):
            pass
        else:
            raise ClickHouseTableExistsError(f'{db}.{table} is exists!')
        if engine_type.lower() == 'view':
            create_sql = f'create view if not exists {db}.{table} as {sql}'
            self.query(create_sql)
        else:
            raise ValueError(f'current not support material view : {engine_type}')

    def create_table_if_not_exists(self, db: str, table: str, sql: (str, pd.DataFrame, None),
                                   key_cols: list,
                                   primary_key_cols=None, sample_expr=None,
                                   engine_type: str = 'ReplacingMergeTree',
                                   extra_format_dict: (dict, None) = None, partitions_expr: (str, None) = None):
        """

        :param db:
        :param table:
        :param sql:
        :param key_cols:
        :param primary_key_cols:
        :param sample_expr:
        :param engine_type:
        :param extra_format_dict:
        :param partitions_expr:
        :return:
        """
        if self._check_exists(f'{db}.{table}', mode='table'):
            pass
        else:
            self._create_table(db, table, sql, key_cols,
                               primary_key_cols=primary_key_cols,
                               sample_expr=sample_expr,
                               engine_type=engine_type,
                               extra_format_dict=extra_format_dict,
                               partitions_expr=partitions_expr)

    def _create_table(self, db: str, table: str, sql: (str, pd.DataFrame, None), key_cols: list,
                      primary_key_cols=None, sample_expr=None,
                      engine_type: str = 'ReplacingMergeTree',
                      extra_format_dict: (dict, None) = None, partitions_expr: (str, None) = None) -> object:

        # if sql is None:  # create table from sql
        #     create_sql = self._create_table_from_sql(db, table, None, key_cols,
        #                                              primary_key_cols=primary_key_cols, sample_expr=sample_expr,
        #                                              engine_type=engine_type,
        #                                              extra_format_dict=extra_format_dict,
        #                                              partitions_expr=partitions_expr,
        #                                              )
        if isinstance(sql, str):
            create_sql = self._create_table_from_sql(db, table, sql, key_cols,
                                                     primary_key_cols=primary_key_cols,
                                                     sample_expr=sample_expr,
                                                     engine_type=engine_type,
                                                     extra_format_dict=extra_format_dict,
                                                     partitions_expr=partitions_expr)
        elif isinstance(sql, pd.DataFrame):
            create_sql = self._create_table_from_df(db, table, sql, key_cols,
                                                    primary_key_cols=primary_key_cols,
                                                    sample_expr=sample_expr,
                                                    engine_type=engine_type,
                                                    extra_format_dict=extra_format_dict,
                                                    partitions_expr=partitions_expr)
        else:
            raise ParameterTypeError(f'unknown sql type: {type(sql)} @ {sql}')

        exist_status = self._check_exists(db_table=f'{db}.{table}')
        if exist_status:
            raise ClickHouseTableExistsError(f'{db}.{table} is exists!')
        self.query(create_sql)

    # def _create_table_from_df(self, db: str, table: str, df: pd.DataFrame, key_cols: (list, tuple),
    #                           primary_key_cols=None, sample_expr=None,
    #                           engine_type: str = 'ReplacingMergeTree', extra_format_dict=None, partitions_expr=None,
    #                           settings="SETTINGS index_granularity = 8192",
    #                           other=''):
    #
    #     df = self.translate_dtypes1_as_dtypes2(df, src2target={'category': 'str'})
    #     cols = df.columns
    #     dtypes_dict = self.translate_dtypes_from_df(df)
    #     if extra_format_dict is None:
    #         pass
    #     else:
    #         dtypes_dict.update(extra_format_dict)
    #     dtypes_dict = {k: v for k, v in dtypes_dict.items() if k in cols}
    #     base = self._create_table_sql(db, table, dtypes_dict, key_cols,
    #                                   primary_key_cols=primary_key_cols,
    #                                   sample_expr=sample_expr,
    #                                   engine_type=engine_type,
    #                                   partitions_expr=partitions_expr,
    #                                   settings=settings,
    #                                   other=other
    #                                   )
    # exist_status = self._check_exists(db_table=f'{db}.{table}')
    # if exist_status:
    #     raise ClickHouseTableExistsError(f'{db}.{table} is exists!')
    # # self.query(base)
    # return base

    def _create_table_from_sql(self, db: str, table: str, sql: str, key_cols: list,
                               extra_format_dict: (dict, None) = None,
                               primary_key_cols=None, sample_expr=None,
                               engine_type: str = 'ReplacingMergeTree',
                               settings="SETTINGS index_granularity = 8192",
                               other='',
                               partitions_expr=None):
        """


        :param db:
        :param table:
        :param sql:
        :param key_cols:
        :param engine_type:
        :param extra_format_dict:
        :return:
        """

        if isinstance(sql, str):
            # if self._check_sql_type(sql) == 'insert-liked':
            if sql.lower().startswith('create'):
                # if sql is create table sql will return directly
                return sql
        else:
            raise ParameterTypeError('sql must be string')
        # detect end with limit
        # limit_status = self._check_end_with_limit(sql, pattern=r'[\s]+limit[\s]+[0-9]+$')
        # if limit_status:
        #     describe_sql = f' describe({sql}) '
        # else:
        #     describe_sql = f'describe ( {sql} limit 1)'
        describe_sql = self._obtain_describe_sql(sql)
        dtypes_df = self.query(describe_sql)
        dtypes_dict = dict(dtypes_df[['name', 'type']].drop_duplicates().values)

        if isinstance(extra_format_dict, dict):
            dtypes_dict.update(extra_format_dict)
        else:
            pass
        sql = self._create_table_sql(db, table, dtypes_dict, key_cols,
                                     primary_key_cols=primary_key_cols,
                                     sample_expr=sample_expr,
                                     engine_type=engine_type,
                                     partitions_expr=partitions_expr,
                                     settings=settings,
                                     other=other)

        # self.query(sql)
        return sql


if __name__ == '__main__':
    pass



# coding=utf-8
import gzip
import warnings
from collections import namedtuple
from urllib import parse

import asyncio
import json
import nest_asyncio
import pandas as pd
import requests
from aiohttp import ClientSession
from functools import partial, partialmethod

from ClickSQL.conf.parse_rfc_1738_args import parse_rfc1738_args
from ClickSQL.errors import ParameterKeyError, ParameterTypeError, DatabaseTypeError, DatabaseError, \
    HeartbeatCheckFailure, ClickHouseTableNotExistsError
from ClickSQL.utils.file_cache import file_cache

"""
this will hold base function of clickhouse and it will apply a path of access clickhouse through clickhouse api service
this scripts will use none of clickhouse client and only depend on requests to make transactions with 
clickhouse-server

"""

nest_asyncio.apply()  # allow run at jupyter and asyncio env

node_parameters = ('host', 'port', 'user', 'password', 'database')
node = namedtuple('clickhouse', node_parameters)
available_queries_select = ('select', 'show', 'desc')
available_queries_insert = ('insert', 'optimize', 'create')
PRINT_CHECK_RESULT = True
GLOBAL_RAISE_ERROR = True
SEMAPHORE = 10  # control async number for whole query list


# class SmartResult(object):
#
#     def __init__(self, result, status_code: int):
#         self._obj = result
#         self.status_code = status_code


def SmartBytes(result: bytes, status_code: int):
    result_cls = type('SmartBytes', (bytes,), {'status_code': property(lambda x: status_code)})

    return result_cls(result)


# TODO change to queue mode change remove aiohttp depends
class ClickHouseTools(object):
    @staticmethod
    def _check_sql_type(sql: str):
        if sql.lower().startswith(available_queries_select):
            return 'select-liked'
        elif sql.lower().startswith(available_queries_insert):
            return 'insert-liked'

        else:
            raise ValueError('sql type cannot be supported! need upgrade!')

    @staticmethod
    def _transfer_sql_format(sql: str, convert_to: str, transfer_sql_format: bool = True):
        """
        provide a method which will translate a standard sql into clickhouse sql with might use format as suffix
        :param sql:
        :param convert_to:
        :param transfer_sql_format:
        :return:
        """
        if transfer_sql_format:
            if convert_to is None:
                clickhouse_format = 'JSON'
            else:
                clickhouse_format = 'JSONCompact' if convert_to.lower() == 'dataframe' else convert_to
            query_with_format = (sql.rstrip('; \n\t') + ' format ' + clickhouse_format).replace('\n', ' ').strip(' ')
            return query_with_format
        else:
            return sql

    @staticmethod
    def _load_into_pd(ret_value: (str, bytes), convert_to: str = 'dataframe', errors='ignore'):
        """
        will provide a approach to load data from clickhouse into pd.DataFrame format which may be easy to use

        :param ret_value:
        :param convert_to:
        :param errors:
        :return:
        """

        if convert_to.lower() == 'dataframe':
            result_dict = json.loads(ret_value, strict=False)
            meta = result_dict['meta']
            name = map(lambda x: x['name'], meta)
            df = pd.DataFrame.from_records(result_dict['data'], columns=list(name))
            for i in meta:
                if i['type'] in ['DateTime', 'Nullable(DateTime)']:  # translate format
                    # process datetime format
                    df[i['name']] = pd.to_datetime(df[i['name']], errors=errors)
            return df
        else:
            return ret_value

    @classmethod
    def _merge_settings(cls, settings: (None, dict), updated_settings: (None, dict) = None,
                        extra_settings: (None, dict) = None):
        """

        :param settings:
        :param updated_settings:
        :return:
        """

        if updated_settings is None:
            updated_settings = {'enable_http_compression': 1, 'send_progress_in_http_headers': 0,
                                'log_queries': 1, 'connect_timeout': 10, 'receive_timeout': 300,
                                'send_timeout': 300, 'output_format_json_quote_64bit_integers': 0,
                                'wait_end_of_query': 0}
        elif not isinstance(updated_settings, dict):
            raise ParameterTypeError(f'updated_settings must be dict type, but get {type(updated_settings)}')
        else:
            pass
        if settings is not None and isinstance(settings, dict):
            invalid_setting_keys = set(settings.keys()) - set(updated_settings.keys())
            if len(invalid_setting_keys) > 0:
                raise ValueError('setting "{0}" are invalid, valid settings are: {1}'.format(
                    ','.join(invalid_setting_keys), ', '.join(updated_settings.keys())))
            updated_settings.update(settings)
        if extra_settings is not None and isinstance(extra_settings, dict):
            updated_settings.update(extra_settings)

        return {k: v * 1 if isinstance(v, bool) else v for k, v in updated_settings.items()}


class ClickHouseBaseNode(ClickHouseTools):
    _default_settings = {'enable_http_compression': 1, 'send_progress_in_http_headers': 0,
                         'log_queries': 1, 'connect_timeout': 10, 'receive_timeout': 300,
                         'send_timeout': 300, 'output_format_json_quote_64bit_integers': 0,
                         'wait_end_of_query': 0}

    __slots__ = ('_db', '_connect_url', '_para', 'http_settings')  # 'max_async_query_once', 'is_closed'

    def __init__(self, **db_settings):
        """

        accepted_formats = ('DataFrame', 'TabSeparated', 'TabSeparatedRaw', 'TabSeparatedWithNames',
                        'TabSeparatedWithNamesAndTypes', 'CSV', 'CSVWithNames', 'Values', 'Vertical', 'JSON',
                        'JSONCompact', 'JSONEachRow', 'TSKV', 'Pretty', 'PrettyCompact',
                        'PrettyCompactMonoBlock', 'PrettyNoEscapes', 'PrettySpace', 'XML')

        :param db_settings:
        """
        self._check_db_settings_(db_settings, available_db_type=[node.__name__])
        self._para = node(db_settings['host'], db_settings['port'], db_settings['user'],
                          db_settings['password'], db_settings['database'])  # store connection information
        self._db = self._para.database
        self._connect_url = 'http://{user}:{passwd}@{host}:{port}'.format(user=self._para.user,
                                                                          passwd=self._para.password,
                                                                          host=self._para.host,
                                                                          port=self._para.port)
        self.http_settings = self._merge_settings(None, updated_settings=self._default_settings,
                                                  extra_settings={'user': self._para.user,
                                                                  'password': self._para.password})
        # self.max_async_query_once = 5
        # self.is_closed = False
        _base_url = "http://{host}:{port}/?".format(host=self._para.host, port=int(self._para.port))
        self.__heartbeat_test__(_base_url)
        self._heartbeat_test_ = partialmethod(self.__heartbeat_test__, _base_url=_base_url)
        self.cache_query = partialmethod(self.execute, enable_cache=True, exploit_func=True)

    @staticmethod
    def _check_db_settings_(db_settings: dict, available_db_type=(node.__name__,)):  # node.__name__ : clickhouse
        """
        it is to check db setting whether is correct!
        :param db_settings:
        :return:
        """

        if isinstance(db_settings, dict):
            if db_settings['name'].lower() not in available_db_type:
                raise DatabaseTypeError(
                    f'database symbol is not accepted, now only accept: {",".join(available_db_type)}')

            missing_keys = filter(lambda x: x not in db_settings.keys(), node_parameters)  # can improve
            if len(tuple(missing_keys)) == 0:
                pass
            else:
                raise ParameterKeyError(f"the following keys are not at settings: {','.join(missing_keys)}")
        else:
            raise ParameterTypeError(f'db_setting must be dict type! but get {type(db_settings)}')

    @staticmethod
    def __heartbeat_test__(_base_url: str):
        """
        a function to test connection by normal way!

        alter function type into staticmethod
        :return:
        """

        ret_value = requests.get(_base_url)
        status_code = ret_value.status_code
        if status_code != 200:
            raise HeartbeatCheckFailure(f'heartbeat check failure at {_base_url}')
        if PRINT_CHECK_RESULT:
            print('connection test: ', ret_value.text.strip())
        del ret_value

    async def _post(self, url: str, sql: str, session, raise_error: bool = True):
        """
        the async way to send post request to the server
        :param url:
        :param sql:
        :param session:
        :return:
        """
        if self.http_settings['enable_http_compression'] == 1:
            async with session.post(url, data=gzip.compress(sql.encode()),
                                    headers={'Content-Encoding': 'gzip',
                                             'Accept-Encoding': 'gzip'}) as resp:
                result = await resp.read()
        else:
            async with session.post(url, body=sql.encode(), ) as resp:
                result = await resp.read()

        result = SmartBytes(result, resp.status)
        # reason = resp.reason
        if result.status_code != 200:
            if raise_error and GLOBAL_RAISE_ERROR:
                raise DatabaseError(result)
            else:
                warnings.warn(str(result))
        return result

    async def _compression_switched_request(self, query_with_format: (tuple, list, str), convert_to: str = 'dataframe',
                                            transfer_sql_format: bool = True, sem=None, raise_error=True):
        """
        the core request operator with compression switch adaptor

        :param query_with_format:
        :param convert_to:
        :param transfer_sql_format:
        :param sem:
        :return:
        """
        url = self._connect_url + '/?' + parse.urlencode(self.http_settings)
        transfer_sql = partial(self._transfer_sql_format, convert_to=convert_to,
                               transfer_sql_format=transfer_sql_format)
        if sem is None:
            sem = asyncio.Semaphore(SEMAPHORE)  # limit async num
        async with sem:  # limit async number
            async with ClientSession() as session:
                if isinstance(query_with_format, str):
                    result = await self._post(url, transfer_sql(query_with_format), session, raise_error=raise_error)
                elif isinstance(query_with_format, (tuple, list)):
                    result = [await self._post(url, transfer_sql(sql), session, raise_error=raise_error) for sql in
                              query_with_format]
                else:
                    raise ValueError('query_with_format must be str , list or tuple')
        return result

    @classmethod
    def _load_into_pd_ext(cls, sql: (str, list, tuple), ret_value: (bytes, list, tuple), convert_to: str,
                          to_df: bool = True):
        """
        a way to parse into dataframe
        :param sql:
        :param ret_value:
        :param convert_to:
        :param to_df:
        :return:
        """
        if not to_df:
            result = ret_value
        elif isinstance(sql, str):
            if ret_value != b'' and ret_value.status_code == 200:
                result = cls._load_into_pd(ret_value, convert_to)
            else:
                result = ret_value
        elif isinstance(sql, (list, tuple)):
            result = [cls._load_into_pd(s, convert_to) if s != b'' and s.status_code == 200 else s for s in
                      ret_value]
        else:
            raise ValueError(f'sql must be str or list or tuple,but get {type(sql)}')
        return result

    def __execute__(self, sql: (str, list, tuple), convert_to: str = 'dataframe', transfer_sql_format: bool = True,
                    loop=None, to_df: bool = True, raise_error=True):
        """
        the core execute function to run the whole requests and SQL or a list of SQL.
        :param sql:
        :param convert_to:
        :param transfer_sql_format:
        :param loop:
        :param to_df:
        :return:
        """

        sem = asyncio.Semaphore(SEMAPHORE)  # limit async num
        resp_list = self._compression_switched_request(sql, convert_to=convert_to,
                                                       transfer_sql_format=transfer_sql_format, sem=sem,
                                                       raise_error=raise_error)
        if loop is None:
            loop = asyncio.get_event_loop()  # init loop
        res = loop.run_until_complete(resp_list)
        result = self._load_into_pd_ext(sql, res, convert_to, to_df=to_df)
        return result

    def execute(self, *sql, convert_to: str = 'dataframe', loop=None, output_df: bool = True,
                enable_cache: bool = False, exploit_func: bool = True, raise_error: bool = True):
        """
        execute sql or multi sql

        :param raise_error:
        :param exploit_func:
        :param enable_cache:
        :param output_df:
        :param sql:
        :param convert_to:
        :param loop:
        :return:
        """
        # TODO change to smart mode, can receive any kind sql combination and handle them
        # detect whether all query are insert process
        # insert_process = map(lambda x: x.lower().startswith(available_queries_insert), sql)
        # # detect whether all query are select process
        # select_process = map(lambda x: x.lower().startswith(available_queries_select), sql)
        # if all(list(select_process)) is True:
        #     to_df = transfer_sql_format = True
        # elif all(list(insert_process)) is True:
        #     to_df = transfer_sql_format = False
        # else:
        #     # TODO change to smart mode, can receive any kind sql combination and handle them
        #     raise ValueError(
        #         'the list of queries must be same type query! currently cannot handle various kind SQL type'
        #         'combination')
        func = file_cache(enable_cache=enable_cache, exploit_func=exploit_func)(self.__execute__)
        result = func(sql, convert_to=convert_to, transfer_sql_format=True, loop=loop,
                      to_df=True * output_df, raise_error=raise_error)

        if len(sql) == 1:
            return result[0]
        else:

            return result

    def __call__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return self.query(*args, **kwargs)

    def query(self, *sql: str, loop=None, output_df: bool = True, raise_error=True):

        """
        add enable_cache and exploit_func

        ## TODO require to upgrade
        :param raise_error:
        :param output_df:
        :param loop:
        :param sql:
        :return:
        """

        result = self.execute(*sql, convert_to='dataframe', loop=loop, output_df=output_df, enable_cache=False,
                              exploit_func=False, raise_error=raise_error)
        return result


class ClickHouseTableNode(ClickHouseBaseNode):

    def __init__(self, conn_str: (str, dict, None) = None, **kwargs):
        """
        add kwargs to contain db settings
        :param conn_str:
        :param kwargs:
        """
        if conn_str is None:
            if kwargs != {}:
                db_settings = kwargs
                db_settings['name'] = node.__name__
            else:
                raise ParameterTypeError('database parameters cannot be parsed normally!')
        elif isinstance(conn_str, str):
            db_settings = parse_rfc1738_args(conn_str)
        elif isinstance(conn_str, dict):
            db_settings = conn_str
        else:
            raise ParameterTypeError(f'conn_str must be str or dict but get: {type(conn_str)}')

        if db_settings['port'] is None:  # add default port for clickhouse
            db_settings['port'] = 8123
        super(ClickHouseTableNode, self).__init__(**db_settings)
        self._table = self.tables[0]

    @property
    def columns(self):
        # db_table: str = None
        db_table = f"{self._db}.{self._table}"
        sql = f'desc {db_table}'
        res = self.execute(sql, convert_to='dataframe')['name'].tolist()
        return res

    @columns.setter
    def columns(self, table: str):
        if table in self.tables:
            self._table = table
        else:
            raise ClickHouseTableNotExistsError(f'{table} not at {self._db}')

    @property
    def tables(self):
        """
        show tables
        :return:
        """
        sql = 'SHOW TABLES FROM {db}'.format(db=self._db)
        res = self.execute(sql, convert_to='dataframe').values.ravel().tolist()
        return res

    @property
    def databases(self):
        """
        show databases
        :return:
        """
        sql = 'SHOW DATABASES'
        res = self.execute(sql, convert_to='dataframe').values.ravel().tolist()
        return res

    def _check_exists(self, db_table: str, mode: str = 'table'):
        """

        :param db_table:
        :param mode:
        :return:
        """
        ## TODO check table exists
        if isinstance(db_table, str):
            if '.' in db_table:
                db, table = db_table.split('.')
            else:
                db, table = self._db, db_table
        else:
            raise ValueError('please input correct db.table information')

        print(f'will detect {db}.{table}')

        if mode == 'table':
            if db == self._db:
                return table in self.tables
            else:
                sql = f"show tables from {db}"
                tables = self.query(sql).values.ravel().tolist()
                return table in tables
            # else:
            # else:
            #     return db in self.databases
        else:

            return db in self.databases


if __name__ == '__main__':
    pass



# # coding=utf-8
# import pandas as pd
# from ClickSQL.utils.lazy_load import LazyInit
#
# from ClickSQL.clickhouse.ClickHouse import ClickHouseTableNode
#
#
# @LazyInit
# class MergeTree(object):
#     def __init__(self, db_table: str, dtypes_dict: dict, order_by_key_cols: (list, tuple), primary_key_cols=None,
#                  sample_expr=None, partitions_expr=None, settings="SETTINGS index_granularity = 8192"):
#         pass
#
#     @classmethod
#     def _create_table_sql(cls, db_table: str, dtypes_dict: dict,
#                           order_by_key_cols: (list, tuple),
#                           primary_key_cols=None, sample_expr=None,
#                           engine_type: str = 'MergeTree', partitions_expr=None,
#                           settings="SETTINGS index_granularity = 8192", other=''):
#         """
#
#         :param db:
#         :param table:
#         :param dtypes_dict:
#         :param order_by_key_cols:
#         :param primary_key_cols:
#         :param sample_expr:
#         :param engine_type:
#         :param partitions_expr:
#         :param settings:
#         :param other:
#         :return:
#         """
#         """CREATE TABLE [IF NOT EXISTS] [db.]table_name [ON CLUSTER cluster]
#                    (
#                        name1 [type1] [DEFAULT|MATERIALIZED|ALIAS expr1],
#                        name2 [type2] [DEFAULT|MATERIALIZED|ALIAS expr2],
#                        ...
#                    ) ENGINE = ReplacingMergeTree([ver])
#                    [PARTITION BY expr]
#                    [ORDER BY expr]
#                    [PRIMARY KEY expr]
#                    [SAMPLE BY expr]
#                    [SETTINGS name=value, ...]"""
#
#         cols_def_str = ''
#         order_by_key_str = ','.join(order_by_key_cols)
#         primary_key_cols_str = ''
#         sample = ''
#         main_body = f"CREATE TABLE IF NOT EXISTS {db_table} ({cols_def_str}) ENGINE={engine_type}"
#         main_settings = f"ORDER BY ({order_by_key_str}) PRIMARY KEY ({primary_key_cols_str}) SAMPLE BY {sample} {settings}"
#
#         cols_def = ','.join([f"{name} {d_type}" for name, d_type in dtypes_dict.items()])
#
#         maid_body = f"CREATE TABLE IF NOT EXISTS {db}.{table} ( {cols_def} ) ENGINE = {engine_type}"
#
#         ORDER_BY_CLAUSE = f"ORDER BY ( {','.join(order_by_key_cols)} )"
#
#         if partitions_expr is None:
#             PARTITION_by_CLAUSE = ''
#         else:
#             PARTITION_by_CLAUSE = f"PARTITION BY {partitions_expr}"
#
#         if primary_key_cols is None:
#             PRIMARY_BY_CLAUSE = ''
#         else:
#             primary_key_expr = ','.join(primary_key_cols)
#             PRIMARY_BY_CLAUSE = f" primary key by ({primary_key_expr})"
#
#         if sample_expr is None:
#             SAMPLE_CLAUSE = ''
#         else:
#             SAMPLE_CLAUSE = sample_expr
#
#         base = f"{maid_body} {PARTITION_by_CLAUSE} {ORDER_BY_CLAUSE} {PRIMARY_BY_CLAUSE} {SAMPLE_CLAUSE} {other} {settings}"
#
#         return base
#
#
# class ClickHouseTableNodeExt(ClickHouseTableNode):
#     @classmethod
#     def _create_table(cls, obj: object, db: str, table: str, sql: str, key_cols: list,
#                       engine_type: str = 'ReplacingMergeTree',
#                       extra_format_dict: (dict, None) = None, partitions_expr: (str, None) = None) -> object:
#         if isinstance(sql, str):
#             return cls._create_table_from_sql(obj, db, table, sql, key_cols,
#                                               engine_type=engine_type,
#                                               extra_format_dict=extra_format_dict, partitions_expr=partitions_expr)
#         elif isinstance(sql, pd.DataFrame):
#             return cls._create_table_from_df(obj, db, table, sql, key_cols,
#                                              engine_type='ReplacingMergeTree',
#                                              extra_format_dict=extra_format_dict, partitions_expr=partitions_expr)
#         else:
#             raise ValueError(f'unknown sql:{sql}')
#
#     @classmethod
#     def _create_table_from_df(cls, obj: object, db: str, table: str, df: pd.DataFrame, key_cols: (list, tuple),
#                               engine_type: str = 'ReplacingMergeTree', extra_format_dict=None, partitions_expr=None):
#         query_func = obj.query
#
#         df = cls.translate_dtypes1_as_dtypes2(df, src2target={'category': 'str'})
#         cols = df.columns
#         dtypes_dict = cls.translate_dtypes_from_df(df)
#         if extra_format_dict is None:
#             pass
#         else:
#             dtypes_dict.update(extra_format_dict)
#         dtypes_dict = {k: v for k, v in dtypes_dict.items() if k in cols}
#         base = cls._create_table_from_sql(db, table, dtypes_dict, key_cols, engine_type=engine_type,
#                                           extra_format_dict=extra_format_dict, partitions_expr=partitions_expr)
#         exist_status = cls._check_table_exists(obj, db, table)
#
#         query_func(base)
#         return exist_status
#
#     @classmethod
#     def _create_table_sql(cls, db: str, table: str, dtypes_dict: dict,
#                           order_by_key_cols: (list, tuple),
#                           primary_key_cols=None, sample_expr=None,
#                           engine_type: str = 'ReplacingMergeTree', partitions_expr=None,
#                           settings="SETTINGS index_granularity = 8192", other=''):
#         """
#
#         :param db:
#         :param table:
#         :param dtypes_dict:
#         :param order_by_key_cols:
#         :param primary_key_cols:
#         :param sample_expr:
#         :param engine_type:
#         :param partitions_expr:
#         :param settings:
#         :param other:
#         :return:
#         """
#         """CREATE TABLE [IF NOT EXISTS] [db.]table_name [ON CLUSTER cluster]
#                    (
#                        name1 [type1] [DEFAULT|MATERIALIZED|ALIAS expr1],
#                        name2 [type2] [DEFAULT|MATERIALIZED|ALIAS expr2],
#                        ...
#                    ) ENGINE = ReplacingMergeTree([ver])
#                    [PARTITION BY expr]
#                    [ORDER BY expr]
#                    [PRIMARY KEY expr]
#                    [SAMPLE BY expr]
#                    [SETTINGS name=value, ...]"""
#
#         cols_def = ','.join([f"{name} {d_type}" for name, d_type in dtypes_dict.items()])
#
#         maid_body = f"CREATE TABLE IF NOT EXISTS {db}.{table} ( {cols_def} ) ENGINE = {engine_type}"
#
#         ORDER_BY_CLAUSE = f"ORDER BY ( {','.join(order_by_key_cols)} )"
#
#         if partitions_expr is None:
#             PARTITION_by_CLAUSE = ''
#         else:
#             PARTITION_by_CLAUSE = f"PARTITION BY {partitions_expr}"
#
#         if primary_key_cols is None:
#             PRIMARY_BY_CLAUSE = ''
#         else:
#             primary_key_expr = ','.join(primary_key_cols)
#             PRIMARY_BY_CLAUSE = f"PARTITION BY ({primary_key_expr})"
#
#         if sample_expr is None:
#             SAMPLE_CLAUSE = ''
#         else:
#             SAMPLE_CLAUSE = sample_expr
#
#         base = f"{maid_body} {PARTITION_by_CLAUSE} {ORDER_BY_CLAUSE} {PRIMARY_BY_CLAUSE} {SAMPLE_CLAUSE} {other} {settings}"
#
#         return base
#
#     @staticmethod
#     def _check_end_with_limit(string, pattern=r'[\s]+limit[\s]+[0-9]+$'):
#         m = re.findall(pattern, string)
#         if m is None or m == []:
#             return False
#         else:
#             return True
#
#     @classmethod
#     def _create_table_from_sql(cls, db: str, table: str, sql: str, key_cols: list,
#                                extra_format_dict: (dict, None) = None,
#                                primary_key_cols=None, sample_expr=None, other='',
#                                engine_type: str = 'ReplacingMergeTree',
#                                partitions_expr=None, query_func=None):
#         """
#
#         :param obj:
#         :param db:
#         :param table:
#         :param sql:
#         :param key_cols:
#         :param engine_type:
#         :param extra_format_dict:
#         :return:
#         """
#
#         if isinstance(sql, str):
#             pass
#         else:
#             raise ValueError('sql must be string')
#
#         limit_status = cls._check_end_with_limit(sql, pattern=r'[\s]+limit[\s]+[0-9]+$')
#         if limit_status:
#             describe_sql = f' describe({sql}) '
#         else:
#             describe_sql = f'describe ( {sql} limit 1)'
#
#         if query_func is None:
#             raise ValueError('query function should be set!')
#
#         dtypes_df = query_func(describe_sql)
#
#         dtypes_dict = dict(dtypes_df[['name', 'type']].drop_duplicates().values)
#         if extra_format_dict is None:
#             pass
#         else:
#             dtypes_dict.update(extra_format_dict)
#         sql = cls._create_table_sql(db, table, dtypes_dict, key_cols, engine_type=engine_type,
#                                     primary_key_cols=primary_key_cols, sample_expr=sample_expr,
#                                     partitions_expr=partitions_expr, other=other)
#
#         return sql
#
#     @staticmethod
#     def translate_dtypes1_as_dtypes2(df: pd.DataFrame, src2target={'category': 'str'}):
#         dtypes_series = df.dtypes
#         for src, dest in src2target.items():
#             if src in dtypes_series:
#                 category_cols = dtypes_series[dtypes_series == src].index
#                 for col in category_cols:
#                     df[col] = df[col].astype(dest)
#             else:
#                 pass
#         return df
#
#     @staticmethod
#     def translate_dtypes_from_df(df: pd.DataFrame, translate_dtypes: dict = {'object': 'String',
#                                                                              'datetime64[ns]': 'Datetime'}):
#         if hasattr(df, 'dtypes'):
#             dtypes_series = df.dtypes.replace(translate_dtypes)
#             return dtypes_series.map(lambda x: str(x).capitalize()).to_dict()
#         elif hasattr(df, '_columns_') and 'type' in df._columns_ and 'name' in df._columns_:
#             dtypes_series = df.set_index('name')['type'].replace(translate_dtypes)
#             return dtypes_series.map(lambda x: str(x)).to_dict()
#         else:
#             raise ValueError(f'unknown df:{type(df)}')
#
#     @classmethod
#     def _create_table_from_df(cls, db: str, table: str, df: pd.DataFrame, key_cols: (list, tuple),
#                               engine_type: str = 'ReplacingMergeTree', extra_format_dict=None, partitions_expr=None,
#                               src2target={'category': 'str'},
#                               query_func=None
#                               ):
#
#         df = cls.translate_dtypes1_as_dtypes2(df, src2target={'category': 'str'})
#         cols = df.columns
#         dtypes_dict = cls.translate_dtypes_from_df(df)
#         if extra_format_dict is None:
#             pass
#         else:
#             dtypes_dict.update(extra_format_dict)
#         dtypes_dict = {k: v for k, v in dtypes_dict.items() if k in cols}
#         base = cls._create_table_from_sql(db, table, dtypes_dict, key_cols, engine_type=engine_type,
#                                           extra_format_dict=extra_format_dict, partitions_expr=partitions_expr)
#         exist_status = cls._check_table_exists(obj, db, table)
#
#         query_func(base)
#         return exist_status
#
#     @classmethod
#     def _check_table_exists(cls, obj, db, table):
#         ## todo check the table exists
#         pass



# coding=utf-8


class ParameterKeyError(Exception):
    pass


class ParameterTypeError(Exception):
    pass


class DatabaseTypeError(Exception):
    pass


class DatabaseNotExists(Exception): pass


class DatabaseError(Exception):
    pass


class HeartbeatCheckFailure(Exception):
    pass


class ClickHouseTableExistsError(Exception):
    pass


class ClickHouseTableNotExistsError(Exception):
    pass
